"""pytest test configuration file."""
# pylint: disable=redefined-outer-name

import json
from pathlib import Path
from typing import Dict, Generator, List
from uuid import UUID

import pytest
import responses
from _pytest.monkeypatch import MonkeyPatch

from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.scryfall.models import (
    ScryCard,
    ScryMigration,
    ScryObjectList,
    ScryRootList,
    ScrySet,
)

TEST_DATA_DIR = Path(__file__).parent / "data"
SETS_DATA_FILE = TEST_DATA_DIR / "sets.json"
CARDS_DATA_FILE = TEST_DATA_DIR / "cards.json"
MIGRATIONS_DATA_FILE = TEST_DATA_DIR / "migrations.json"


@pytest.fixture(autouse=True)
def fetcher_cache_dir(tmp_path: Path, monkeypatch: MonkeyPatch) -> Path:
    """Patch fetcher cache dirs for testing."""
    cache_path = tmp_path / "cache"
    monkeypatch.setattr("mtg_ssm.scryfall.fetcher.CACHE_DIR", cache_path)
    return cache_path


@pytest.fixture(autouse=True)
def requests_mock() -> Generator[responses.RequestsMock, None, None]:
    """Auto replace all requests with a mock."""
    with responses.RequestsMock() as rsps:
        yield rsps


@pytest.fixture(scope="session")
def cards_data() -> List[ScryCard]:
    """Fixture containing all test card data."""
    with CARDS_DATA_FILE.open("rt", encoding="utf-8") as card_data_file:
        card_json = json.load(card_data_file)
    return ScryRootList[ScryCard].parse_obj(card_json).__root__


@pytest.fixture(scope="session")
def sets_data() -> List[ScrySet]:
    """Fixture containing all test set data."""
    with SETS_DATA_FILE.open("rt", encoding="utf-8") as sets_data_file:
        sets_json = json.load(sets_data_file)
    return ScryObjectList[ScrySet].parse_obj(sets_json).data


@pytest.fixture(scope="session")
def migrations_data() -> List[ScryMigration]:
    """Fixture containing all test migrations data."""
    with MIGRATIONS_DATA_FILE.open("rt", encoding="utf-8") as migrations_data_file:
        migrations_json = json.load(migrations_data_file)
    return ScryObjectList[ScryMigration].parse_obj(migrations_json).data


@pytest.fixture(scope="session")
def id_to_card(cards_data: List[ScryCard]) -> Dict[UUID, ScryCard]:
    """Fixture returning scryfall id to card object for all test card data."""
    return {card.id: card for card in cards_data}


@pytest.fixture(scope="session")
def scryfall_data(
    cards_data: List[ScryCard],
    sets_data: List[ScrySet],
    migrations_data: List[ScryMigration],
) -> ScryfallDataSet:
    """Fixture containing all scryfall test data."""
    return ScryfallDataSet(sets=sets_data, cards=cards_data, migrations=migrations_data)



"""Scryfall object models."""

import datetime as dt
from decimal import Decimal
from enum import Enum
from typing import Dict, Generic, List, Literal, Optional, TypeVar, Union
from uuid import UUID

from pydantic import BaseModel, HttpUrl
from pydantic.generics import GenericModel


class ScryColor(str, Enum):
    """Enum for https://scryfall.com/docs/api/colors#color-arrays"""

    WHITE = "W"
    BLUE = "U"
    BLACK = "B"
    RED = "R"
    GREEN = "G"
    COLORLESS = "C"
    TAP = "T"


class ScrySetType(str, Enum):
    """Enum for https://scryfall.com/docs/api/sets#set-types"""

    CORE = "core"
    EXPANSION = "expansion"
    MASTERS = "masters"
    MASTERPIECE = "masterpiece"
    FROM_THE_VAULT = "from_the_vault"
    SPELLBOOK = "spellbook"
    PREMIUM_DECK = "premium_deck"
    DUEL_DECK = "duel_deck"
    DRAFT_INNOVATION = "draft_innovation"
    TREASURE_CHEST = "treasure_chest"
    COMMANDER = "commander"
    PLANECHASE = "planechase"
    ARCHENEMY = "archenemy"
    VANGUARD = "vanguard"
    FUNNY = "funny"
    STARTER = "starter"
    BOX = "box"
    PROMO = "promo"
    TOKEN = "token"
    MEMORABILIA = "memorabilia"
    ALCHEMY = "alchemy"
    ARSENAL = "arsenal"


class ScryCardLayout(str, Enum):
    """Enum for https://scryfall.com/docs/api/layouts#layout"""

    NORMAL = "normal"
    SPLIT = "split"
    FLIP = "flip"
    TRANSFORM = "transform"
    MODAL_DFC = "modal_dfc"
    MELD = "meld"
    LEVELER = "leveler"
    CLASS = "class"
    SAGA = "saga"
    ADVENTURE = "adventure"
    PLANAR = "planar"
    SCHEME = "scheme"
    VANGUARD = "vanguard"
    TOKEN = "token"
    DOUBLE_FACED_TOKEN = "double_faced_token"
    EMBLEM = "emblem"
    AUGMENT = "augment"
    HOST = "host"
    ART_SERIES = "art_series"
    DOUBLE_SIDED = "double_sided"
    REVERSIBLE_CARD = "reversible_card"


class ScryCardFrame(str, Enum):
    """Enum for https://scryfall.com/docs/api/layouts#frames"""

    Y1993 = "1993"
    Y1997 = "1997"
    Y2003 = "2003"
    Y2015 = "2015"
    FUTURE = "future"


class ScryFrameEffect(str, Enum):
    """Enum for https://scryfall.com/docs/api/layouts#frame-effects"""

    NONE = ""
    LEGENDARY = "legendary"
    MIRACLE = "miracle"
    NYXBORN = "nyxborn"
    NYXTOUCHED = "nyxtouched"
    DRAFT = "draft"
    DEVOID = "devoid"
    TOMBSTONE = "tombstone"
    COLORSHIFTED = "colorshifted"
    INVERTED = "inverted"
    SUNMOONDFC = "sunmoondfc"
    COMPASSLANDDFC = "compasslanddfc"
    ORIGINPWDFC = "originpwdfc"
    MOONELDRAZIDFC = "mooneldrazidfc"
    MOONREVERSEMOONDFC = "moonreversemoondfc"
    WAXINGANDWANINGMOONDFC = "waxingandwaningmoondfc"
    SHOWCASE = "showcase"
    EXTENDEDART = "extendedart"
    COMPANION = "companion"
    FULLART = "fullart"
    ETCHED = "etched"
    SNOW = "snow"
    LESSON = "lesson"
    TEXTLESS = "textless"
    SHATTEREDGLASS = "shatteredglass"
    CONVERTDFC = "convertdfc"
    FANDFC = "fandfc"
    UPSIDEDOWNDFC = "upsidedowndfc"
    GILDED = "gilded"


class ScryBorderColor(str, Enum):
    """Enum for card border_color"""

    BLACK = "black"
    BORDERLESS = "borderless"
    GOLD = "gold"
    SILVER = "silver"
    WHITE = "white"


class ScryFinish(str, Enum):
    """Enum for card finishes"""

    FOIL = "foil"
    NONFOIL = "nonfoil"
    ETCHED = "etched"
    GLOSSY = "glossy"


class ScryImageStatus(str, Enum):
    """Enum for card image_status"""

    MISSING = "missing"
    PLACEHOLDER = "placeholder"
    LOWRES = "lowres"
    HIGHRES_SCAN = "highres_scan"


class ScryGame(str, Enum):
    """Enum for card games"""

    PAPER = "paper"
    ARENA = "arena"
    MTGO = "mtgo"
    SEGA = "sega"
    ASTRAL = "astral"


class ScryRarity(str, Enum):
    """Enum for card rarity"""

    COMMON = "common"
    UNCOMMON = "uncommon"
    RARE = "rare"
    MYTHIC = "mythic"
    SPECIAL = "special"
    BONUS = "bonus"


class ScrySecurityStamp(str, Enum):
    """Enum for card security_stamp"""

    OVAL = "oval"
    TRIANGLE = "triangle"
    ACORN = "acorn"
    ARENA = "arena"
    CIRCLE = "circle"
    HEART = "heart"


class ScryFormat(str, Enum):
    """Enum for card legalities keys"""

    BRAWL = "brawl"
    COMMANDER = "commander"
    DUEL = "duel"
    FRONTIER = "frontier"
    FUTURE = "future"
    LEGACY = "legacy"
    MODERN = "modern"
    OLDSCHOOL = "oldschool"
    PAUPER = "pauper"
    PENNY = "penny"
    STANDARD = "standard"
    VINTAGE = "vintage"
    HISTORIC = "historic"
    PIONEER = "pioneer"
    GLADIATOR = "gladiator"
    EXPLORER = "explorer"
    HISTORICBRAWL = "historicbrawl"
    ALCHEMY = "alchemy"
    PAUPERCOMMANDER = "paupercommander"
    PREMODERN = "premodern"


class ScryLegality(str, Enum):
    """Enum for card legalities values"""

    LEGAL = "legal"
    NOT_LEGAL = "not_legal"
    RESTRICTED = "restricted"
    BANNED = "banned"


class ScryMigrationStrategy(str, Enum):
    """Enum for migration strategy values"""

    MERGE = "merge"
    DELETE = "delete"


T = TypeVar("T")


class ScryRootList(GenericModel, Generic[T]):
    """Model for unstructured list of scryfall objects (e.g. bulk file data)"""

    __root__: List[T]


class ScryObjectList(GenericModel, Generic[T]):
    """Model for https://scryfall.com/docs/api/lists"""

    object: Literal["list"] = "list"
    data: List[T]
    has_more: bool
    next_page: Optional[HttpUrl]
    total_cards: Optional[int]
    warnings: Optional[List[str]]


class ScrySet(BaseModel):
    """Model for https://scryfall.com/docs/api/sets"""

    object: Literal["set"] = "set"
    id: UUID
    code: str
    mtgo_code: Optional[str]
    arena_code: Optional[str]
    tcgplayer_id: Optional[int]
    name: str
    set_type: ScrySetType
    released_at: Optional[dt.date]
    block_code: Optional[str]
    block: Optional[str]
    parent_set_code: Optional[str]
    card_count: int
    printed_size: Optional[int]
    digital: bool
    foil_only: bool
    nonfoil_only: Optional[bool]
    icon_svg_uri: HttpUrl
    search_uri: HttpUrl
    scryfall_uri: HttpUrl
    uri: HttpUrl


class ScryRelatedCard(BaseModel):
    """Model for https://scryfall.com/docs/api/cards#related-card-objects"""

    object: Literal["related_card"] = "related_card"
    id: UUID
    component: str
    name: str
    type_line: str
    uri: HttpUrl


class ScryCardFace(BaseModel):
    """Model for https://scryfall.com/docs/api/cards#card-face-objects"""

    object: Literal["card_face"] = "card_face"
    artist: Optional[str]
    artist_id: Optional[UUID]
    cmc: Optional[Decimal]
    color_indicator: Optional[List[ScryColor]]
    colors: Optional[List[ScryColor]]
    flavor_name: Optional[str]
    flavor_text: Optional[str]
    illustration_id: Optional[UUID]
    image_uris: Optional[Dict[str, HttpUrl]]
    layout: Optional[ScryCardLayout]
    loyalty: Optional[str]
    mana_cost: str
    name: str
    oracle_id: Optional[UUID]
    oracle_text: Optional[str]
    power: Optional[str]
    printed_name: Optional[str]
    printed_text: Optional[str]
    printed_type_line: Optional[str]
    toughness: Optional[str]
    type_line: Optional[str]
    watermark: Optional[str]


class CardPreviewBlock(BaseModel):
    """Model for card preview block."""

    source: str
    source_uri: Union[HttpUrl, Literal[""], str]
    previewed_at: dt.date


class ScryCard(BaseModel):
    """Model for https://scryfall.com/docs/api/cards"""

    object: Literal["card"] = "card"
    # Core Card Fields
    arena_id: Optional[int]
    id: UUID
    lang: str
    mtgo_id: Optional[int]
    mtgo_foil_id: Optional[int]
    multiverse_ids: Optional[List[int]]
    tcgplayer_id: Optional[int]
    tcgplayer_etched_id: Optional[int]
    cardmarket_id: Optional[int]
    oracle_id: Optional[UUID]
    prints_search_uri: HttpUrl
    rulings_uri: HttpUrl
    scryfall_uri: HttpUrl
    uri: HttpUrl
    # Gameplay Fields
    all_parts: Optional[List[ScryRelatedCard]]
    card_faces: Optional[List[ScryCardFace]]
    cmc: Optional[Decimal]
    colors: Optional[List[ScryColor]]
    color_identity: List[ScryColor]
    color_indicator: Optional[List[ScryColor]]
    edhrec_rank: Optional[int]
    foil: bool
    hand_modifier: Optional[str]
    keywords: List[str]
    layout: ScryCardLayout
    legalities: Dict[ScryFormat, ScryLegality]
    life_modifier: Optional[str]
    loyalty: Optional[str]
    mana_cost: Optional[str]
    name: str
    nonfoil: bool
    oracle_text: Optional[str]
    oversized: bool
    penny_rank: Optional[int]
    power: Optional[str]
    produced_mana: Optional[List[Union[ScryColor, int]]]
    reserved: bool
    toughness: Optional[str]
    type_line: Optional[str]
    # Print Fields
    artist: Optional[str]
    artist_ids: Optional[List[UUID]]
    booster: bool
    border_color: ScryBorderColor
    card_back_id: Optional[UUID]
    collector_number: str
    content_warning: Optional[bool]
    digital: bool
    finishes: List[ScryFinish]
    flavor_name: Optional[str]
    flavor_text: Optional[str]
    frame_effect: Optional[ScryFrameEffect]
    frame_effects: Optional[List[ScryFrameEffect]]
    frame: ScryCardFrame
    full_art: bool
    games: List[ScryGame]
    highres_image: bool
    illustration_id: Optional[UUID]
    image_status: ScryImageStatus
    image_uris: Optional[Dict[str, HttpUrl]]
    prices: Optional[Dict[str, Optional[Decimal]]]  # TODO: enum keys
    printed_name: Optional[str]
    printed_text: Optional[str]
    printed_type_line: Optional[str]
    promo: bool
    promo_types: Optional[List[str]]
    purchase_uris: Optional[Dict[str, HttpUrl]]
    rarity: ScryRarity
    related_uris: Optional[Dict[str, HttpUrl]]
    released_at: dt.date
    reprint: bool
    scryfall_set_uri: HttpUrl
    set_name: str
    set_search_uri: HttpUrl
    set_type: str
    set_uri: HttpUrl
    set: str
    set_id: UUID
    story_spotlight: bool
    textless: bool
    variation: bool
    variation_of: Optional[UUID]
    security_stamp: Optional[ScrySecurityStamp]
    watermark: Optional[str]
    preview: Optional[CardPreviewBlock]


class ScryBulkData(BaseModel):
    """Model for https://scryfall.com/docs/api/bulk-data"""

    object: Literal["bulk_data"] = "bulk_data"
    id: UUID
    uri: HttpUrl
    type: str
    name: str
    description: str
    download_uri: HttpUrl
    updated_at: dt.datetime
    compressed_size: Optional[int]
    content_type: str
    content_encoding: str


class ScryMigration(BaseModel):
    """Model for https://scryfall.com/docs/api/migrations"""

    object: Literal["migration"] = "migration"
    id: UUID
    uri: HttpUrl
    performed_at: dt.date
    migration_strategy: ScryMigrationStrategy
    old_scryfall_id: UUID
    new_scryfall_id: Optional[UUID]
    note: Optional[str]



"""Interface definition for serializers."""

import abc
from pathlib import Path
from typing import ClassVar, Dict, List, Optional, Set, Tuple, Type

from mtg_ssm.containers.collection import MagicCollection
from mtg_ssm.containers.indexes import Oracle


class Error(Exception):
    """Base error for serializers."""


class UnknownDialect(Exception):
    """Raised when an (extension, dialect) pair is requested."""


class DeserializationError(Error):
    """Raised when there is an error reading counts from a file."""


class SerializationDialect(metaclass=abc.ABCMeta):
    """Abstract interface for mtg ssm serialization dialect."""

    _EXT_DIALECT_DOC: ClassVar[Set[Tuple[str, str, str]]] = set()
    _EXT_DIALECT_TO_IMPL: ClassVar[
        Dict[Tuple[str, str], Type["SerializationDialect"]]
    ] = {}

    extension: ClassVar[Optional[str]] = None
    dialect: ClassVar[Optional[str]] = None

    def __init_subclass__(cls: Type["SerializationDialect"]) -> None:
        super().__init_subclass__()
        if cls.extension is not None and cls.dialect is not None:
            cls._EXT_DIALECT_DOC.add(
                (cls.extension, cls.dialect, cls.__doc__ or cls.__name__)
            )
            cls._EXT_DIALECT_TO_IMPL[(cls.extension, cls.dialect)] = cls

    @abc.abstractmethod
    def write(self, path: Path, collection: MagicCollection) -> None:
        """Write print counts to a file."""

    @abc.abstractmethod
    def read(self, path: Path, oracle: Oracle) -> MagicCollection:
        """Read print counts from file."""

    @classmethod
    def dialects(
        cls: Type["SerializationDialect"],
    ) -> List[Tuple[str, Optional[str], Optional[str]]]:
        """List of (extension, dialect, description) of registered dialects."""
        return sorted(
            (ext, dial or "", doc or "") for ext, dial, doc in cls._EXT_DIALECT_DOC
        )

    @classmethod
    def by_extension(
        cls: Type["SerializationDialect"],
        extension: str,
        dialect_mappings: Dict[str, str],
    ) -> Type["SerializationDialect"]:
        """Get a serializer class for a given extension and dialect mapping."""
        dialect = dialect_mappings.get(extension, extension)
        try:
            return cls._EXT_DIALECT_TO_IMPL[(extension, dialect)]
        except KeyError as err:
            raise UnknownDialect(
                f'File extension: "{extension}" dialect: "{dialect}" not found in registry'
            ) from err



"""Scryfall data fetcher."""

import gzip
import json
import os
import pickle
import pprint
import uuid
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Any, Dict, List, Mapping, Union, cast

import appdirs
import requests
from pydantic import ValidationError

from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.scryfall.models import (
    ScryBulkData,
    ScryCard,
    ScryMigration,
    ScryObjectList,
    ScrySet,
)

DEBUG = 0
try:
    DEBUG = int(os.getenv("DEBUG", "0"))
except ValueError:
    pass

APP_AUTHOR = "gwax"
APP_NAME = "mtg_ssm"
CACHE_DIR = Path(appdirs.user_cache_dir(APP_NAME, APP_AUTHOR))

BULK_DATA_ENDPOINT = "https://api.scryfall.com/bulk-data"
SETS_ENDPOINT = "https://api.scryfall.com/sets"
MIGRATIONS_ENDPOINT = "https://api.scryfall.com/migrations"
BULK_TYPE = "default_cards"
OBJECT_CACHE_URL = "file:///$CACHE/pickled_object?v2"

CHUNK_SIZE = 8 * 1024 * 1024
DESERIALIZE_BATCH_SIZE = 50
REQUESTS_TIMEOUT_SECONDS = 10

JSON = Union[str, int, float, bool, None, Mapping[str, Any], List[Any]]


def _value_from_validation_error(data: JSON, verr: ValidationError) -> Dict[str, Any]:
    values = {}
    for error in verr.errors():
        loc = error["loc"]
        value = data
        for field in loc:
            if field == "__root__":
                break
            if isinstance(value, Mapping) and isinstance(field, str):
                value = value[field]
            if isinstance(value, List) and isinstance(field, int):
                value = value[field]
        values[".".join([str(location) for location in loc])] = value
    return values


def _cache_path(endpoint: str, extension: str) -> Path:
    if not extension.startswith("."):
        extension = "." + extension
    cache_id = uuid.uuid5(uuid.NAMESPACE_URL, endpoint)
    return CACHE_DIR / f"{cache_id}{extension}"


def _fetch_endpoint(endpoint: str, *, dirty: bool, write_cache: bool = True) -> JSON:
    CACHE_DIR.mkdir(parents=True, exist_ok=True)
    cache_path = _cache_path(endpoint, ".json.gz")
    if not cache_path.exists():
        dirty = True
    if dirty:
        print(f"Fetching {endpoint}")
        response = requests.get(endpoint, stream=True, timeout=REQUESTS_TIMEOUT_SECONDS)
        response.raise_for_status()
        if not write_cache:
            return response.json()
        print(f"Caching {endpoint}")
        with gzip.open(cache_path, "wb", compresslevel=1) as cache_file:
            for chunk in response.iter_content(chunk_size=CHUNK_SIZE):
                cache_file.write(chunk)
    else:
        print(f"Reading cached {endpoint}")

    with gzip.open(cache_path, "rt", encoding="utf-8") as cache_file:
        return json.load(cache_file)


def _deserialize_cards(card_jsons: List[JSON]) -> List[ScryCard]:
    cards_data: List[ScryCard]
    if DEBUG:
        print("Process pool disabled")
        cards_data = []
        for card_json in card_jsons:
            try:
                cards_data.append(ScryCard.parse_obj(card_json))
            except ValidationError as err:
                print("Failed with pydantic errors on values:")
                pprint.pp(_value_from_validation_error(card_json, err))
                print("Failed on:")
                pprint.pp(card_json)
                raise
            except Exception:
                print("Failed on:")
                pprint.pp(card_json)
                raise
    else:
        with ProcessPoolExecutor() as executor:
            cards_futures = executor.map(
                ScryCard.parse_obj, card_jsons, chunksize=DESERIALIZE_BATCH_SIZE
            )
            cards_data = list(cards_futures)
    return cards_data


def scryfetch() -> ScryfallDataSet:  # pylint: disable=too-many-locals
    """Retrieve and deserialize Scryfall object data."""
    cached_bulk_json = None
    if _cache_path(BULK_DATA_ENDPOINT, ".json.gz").exists():
        cached_bulk_json = _fetch_endpoint(BULK_DATA_ENDPOINT, dirty=False)
    bulk_json = _fetch_endpoint(BULK_DATA_ENDPOINT, dirty=True, write_cache=False)
    cache_dirty = bulk_json != cached_bulk_json

    object_cache_path = _cache_path(OBJECT_CACHE_URL, ".pickle.gz")
    if object_cache_path.exists():
        if cache_dirty or DEBUG:
            object_cache_path.unlink()
        else:
            print("Loading cached scryfall data objects")
            try:
                with gzip.open(object_cache_path, "rb") as object_cache:
                    loaded_data = pickle.load(object_cache)
            except (OSError, pickle.UnpicklingError):
                pass
            else:
                if isinstance(loaded_data, ScryfallDataSet):
                    return loaded_data
            print("Error reading object cache, falling back")

    sets_list = ScryObjectList[ScrySet].parse_obj(
        _fetch_endpoint(SETS_ENDPOINT, dirty=cache_dirty)
    )
    sets_data = sets_list.data
    while sets_list.has_more and sets_list.next_page is not None:
        sets_list = ScryObjectList[ScrySet].parse_obj(
            _fetch_endpoint(sets_list.next_page, dirty=cache_dirty)
        )
        sets_data += sets_list.data

    migrations_list = ScryObjectList[ScryMigration].parse_obj(
        _fetch_endpoint(MIGRATIONS_ENDPOINT, dirty=cache_dirty)
    )
    migrations_data = migrations_list.data
    while migrations_list.has_more and migrations_list.next_page is not None:
        migrations_list = ScryObjectList[ScryMigration].parse_obj(
            _fetch_endpoint(migrations_list.next_page, dirty=cache_dirty)
        )
        migrations_data += migrations_list.data

    bulk_list = ScryObjectList[ScryBulkData].parse_obj(bulk_json)
    bulk_data = bulk_list.data
    [cards_endpoint] = [bd.download_uri for bd in bulk_data if bd.type == BULK_TYPE]

    cards_json = cast(List[JSON], _fetch_endpoint(cards_endpoint, dirty=cache_dirty))

    _fetch_endpoint(BULK_DATA_ENDPOINT, dirty=cache_dirty, write_cache=True)

    print("Deserializing cards")
    cards_data = _deserialize_cards(cards_json)

    scryfall_data = ScryfallDataSet(
        sets=sets_data, cards=cards_data, migrations=migrations_data
    )
    with gzip.open(object_cache_path, "wb", compresslevel=1) as object_cache:
        pickle.dump(scryfall_data, object_cache, protocol=pickle.HIGHEST_PROTOCOL)
    return scryfall_data



"""Tests for mtg_ssm.mtg.counts"""
# pylint: disable=redefined-outer-name

from typing import Any, Dict, List
from uuid import UUID

import pytest

from mtg_ssm.containers import counts
from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.containers.counts import CardNotFoundError, CountType, ScryfallCardCount
from mtg_ssm.containers.indexes import Oracle
from mtg_ssm.containers.legacy import NoMatchError


@pytest.fixture(scope="session")
def oracle(scryfall_data: ScryfallDataSet) -> Oracle:
    """Oracle fixture."""
    return Oracle(scryfall_data)


@pytest.mark.parametrize(
    "in_card_counts, out_card_count",
    [
        pytest.param([], {}, id="no inputs"),
        pytest.param(
            [{UUID(int=1): {CountType.NONFOIL: 2}}],
            {UUID(int=1): {CountType.NONFOIL: 2}},
            id="single input",
        ),
        pytest.param(
            [
                {UUID(int=1): {CountType.NONFOIL: 2}},
                {UUID(int=1): {CountType.NONFOIL: 1, CountType.FOIL: 4}},
            ],
            {UUID(int=1): {CountType.NONFOIL: 3, CountType.FOIL: 4}},
            id="mixed types",
        ),
        pytest.param(
            [
                {UUID(int=1): {CountType.NONFOIL: 2}},
                {UUID(int=1): {CountType.FOIL: 1}, UUID(int=2): {CountType.NONFOIL: 3}},
                {UUID(int=1): {CountType.FOIL: 5}},
            ],
            {
                UUID(int=1): {CountType.NONFOIL: 2, CountType.FOIL: 6},
                UUID(int=2): {CountType.NONFOIL: 3},
            },
            id="multiple inputs",
        ),
    ],
)
def test_merge_card_counts(
    in_card_counts: List[ScryfallCardCount], out_card_count: ScryfallCardCount
) -> None:
    assert counts.merge_card_counts(*in_card_counts) == out_card_count


@pytest.mark.parametrize(
    "left, right, output",
    [
        pytest.param({}, {}, {}, id="no inputs"),
        pytest.param(
            {UUID(int=1): {CountType.NONFOIL: 2}},
            {UUID(int=1): {CountType.NONFOIL: 1}},
            {UUID(int=1): {CountType.NONFOIL: 1}},
            id="positive output",
        ),
        pytest.param(
            {UUID(int=1): {CountType.NONFOIL: 1}},
            {UUID(int=1): {CountType.NONFOIL: 2}},
            {UUID(int=1): {CountType.NONFOIL: -1}},
            id="negative output",
        ),
        pytest.param(
            {UUID(int=1): {CountType.NONFOIL: 1}},
            {UUID(int=1): {CountType.NONFOIL: 1}},
            {},
            id="negated",
        ),
        pytest.param(
            {UUID(int=1): {CountType.NONFOIL: 1}},
            {UUID(int=2): {CountType.NONFOIL: 1}},
            {UUID(int=1): {CountType.NONFOIL: 1}, UUID(int=2): {CountType.NONFOIL: -1}},
            id="mixed cards",
        ),
        pytest.param(
            {UUID(int=1): {CountType.NONFOIL: 1}},
            {UUID(int=1): {CountType.FOIL: 1}},
            {UUID(int=1): {CountType.NONFOIL: 1, CountType.FOIL: -1}},
            id="mixed count types",
        ),
    ],
)
def test_diff_card_counts(
    left: ScryfallCardCount, right: ScryfallCardCount, output: ScryfallCardCount
) -> None:
    assert counts.diff_card_counts(left, right) == output


@pytest.mark.parametrize(
    "card_rows, output",
    [
        pytest.param([], {}, id="nothing"),
        pytest.param([{}], {}, id="empty"),
        pytest.param([{"foo": "bar"}], {}, id="no count"),
        pytest.param(
            [{"foo": "bar", "foil": "1"}],
            {},
            marks=pytest.mark.xfail(raises=NoMatchError),
            id="no id",
        ),
        pytest.param(
            [{"scryfall_id": UUID("9d26f171-5bb6-463c-8473-53b6cc27ed66"), "foil": 1}],
            {UUID("9d26f171-5bb6-463c-8473-53b6cc27ed66"): {counts.CountType.FOIL: 1}},
            id="id and int",
        ),
        pytest.param(
            [{"scryfall_id": "9d26f171-5bb6-463c-8473-53b6cc27ed66", "foil": "1"}],
            {UUID("9d26f171-5bb6-463c-8473-53b6cc27ed66"): {counts.CountType.FOIL: 1}},
            id="text and text",
        ),
        pytest.param(
            [{"scryfall_id": "00000000-0000-0000-0000-000000000001", "nonfoil": "1"}],
            {},
            marks=pytest.mark.xfail(raises=CardNotFoundError),
            id="count with bad id",
        ),
        pytest.param(
            [{"scryfall_id": "00000000-0000-0000-0000-000000000001", "nonfoil": ""}],
            {},
            id="no count with bad id",
        ),
        pytest.param(
            [
                {
                    "scryfall_id": "9d26f171-5bb6-463c-8473-53b6cc27ed66",
                    "foil": "1",
                    "nonfoil": "",
                }
            ],
            {UUID("9d26f171-5bb6-463c-8473-53b6cc27ed66"): {counts.CountType.FOIL: 1}},
            id="empty string",
        ),
        pytest.param(
            [
                {"scryfall_id": "9d26f171-5bb6-463c-8473-53b6cc27ed66", "foil": "1"},
                {"scryfall_id": "758abd53-6ad2-406e-8615-8e48678405b4", "foil": "0"},
                {"scryfall_id": "0180d9a8-992c-4d55-8ac4-33a587786993", "nonfoil": "1"},
            ],
            {
                UUID("9d26f171-5bb6-463c-8473-53b6cc27ed66"): {
                    counts.CountType.FOIL: 1
                },
                UUID("0180d9a8-992c-4d55-8ac4-33a587786993"): {
                    counts.CountType.NONFOIL: 1
                },
            },
            id="multiple",
        ),
        pytest.param(
            [
                {
                    "scryfall_id": UUID("9d26f171-5bb6-463c-8473-53b6cc27ed66"),
                    "foil": 1,
                },
                {
                    "scryfall_id": UUID("9d26f171-5bb6-463c-8473-53b6cc27ed66"),
                    "nonfoil": 1,
                },
                {
                    "scryfall_id": UUID("9d26f171-5bb6-463c-8473-53b6cc27ed66"),
                    "foil": 1,
                },
            ],
            {
                UUID("9d26f171-5bb6-463c-8473-53b6cc27ed66"): {
                    counts.CountType.FOIL: 2,
                    counts.CountType.NONFOIL: 1,
                }
            },
            id="duplicates",
        ),
        pytest.param(
            [
                {
                    "scryfall_id": UUID("585fa2cc-4f77-47ab-8d2c-c68258ced283"),
                    "foil": 1,
                },
            ],
            {
                UUID("9052f5c7-ee3b-457d-97ca-ac6b4518997c"): {
                    counts.CountType.FOIL: 1,
                }
            },
            id="migration",
        ),
    ],
)
def test_aggregate_card_counts(
    oracle: Oracle, card_rows: List[Dict[str, Any]], output: counts.ScryfallCardCount
) -> None:
    assert counts.aggregate_card_counts(card_rows, oracle) == output



"""Tests for mtg_ssm.mtg.util."""

from typing import Dict, Optional
from uuid import UUID

import pytest

from mtg_ssm.mtg import util
from mtg_ssm.scryfall.models import ScryCard


@pytest.mark.parametrize(
    "name, expected",
    [
        pytest.param("Forest", True),
        pytest.param("Snow-Covered Forest", False),
        pytest.param("Wastes", False),
        pytest.param("Abattoir Ghoul", False),
    ],
)
def test_is_strict_basic(name: str, expected: bool) -> None:
    assert util.is_strict_basic(name) is expected


@pytest.mark.parametrize(
    "name, card_id, number, variant",
    [
        pytest.param("Thallid", UUID("4caaf31b-86a9-485b-8da7-d5b526ed1233"), 74, "a"),
        pytest.param(
            "Dark Ritual", UUID("ebb6664d-23ca-456e-9916-afcd6f26aa7f"), 98, None
        ),
        pytest.param(
            "Stairs to Infinity", UUID("57f25ead-b3ec-4c40-972d-d750ed2f5319"), 1, "P"
        ),
        pytest.param(
            "Ertai, the Corrupted",
            UUID("66b950d9-8fef-4deb-b51b-26edb90abc56"),
            107,
            None,
        ),
        pytest.param(
            "Ertai, the Corrupted",
            UUID("fbbfeb32-1654-4bf6-9a38-891f1a03e02b"),
            107,
            "★",
        ),
        pytest.param(
            "Viscera Seer",
            UUID("dba1cf83-e13d-401e-b76f-b12a51b307f9"),
            None,
            "VS",
        ),
    ],
)
def test_collector_int_var(
    id_to_card: Dict[UUID, ScryCard],
    name: str,
    card_id: UUID,
    number: Optional[int],
    variant: Optional[str],
) -> None:
    card = id_to_card[card_id]
    assert card.name == name
    assert (number, variant) == util.collector_int_var(card)



"""Tests for mtg_ssm.serialization.csv."""
# pylint: disable=redefined-outer-name

import textwrap
from pathlib import Path
from typing import Dict
from uuid import UUID

import pytest
from pytest_snapshot.plugin import Snapshot

from mtg_ssm.containers import counts
from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.containers.collection import MagicCollection
from mtg_ssm.containers.counts import ScryfallCardCount
from mtg_ssm.containers.indexes import Oracle
from mtg_ssm.scryfall.models import ScryCard
from mtg_ssm.serialization import csv

TEST_CARD_ID = UUID("57f25ead-b3ec-4c40-972d-d750ed2f5319")


@pytest.fixture(scope="session")
def oracle(scryfall_data: ScryfallDataSet) -> Oracle:
    """Test fixture Oracle with limited data."""
    accepted_sets = {"phop", "pmbs"}
    scryfall_data2 = ScryfallDataSet(
        sets=[s for s in scryfall_data.sets if s.code in accepted_sets],
        cards=[c for c in scryfall_data.cards if c.set in accepted_sets],
        migrations=[],
    )
    return Oracle(scryfall_data2)


def test_header() -> None:
    assert csv.CSV_HEADER == [
        "set",
        "name",
        "collector_number",
        "scryfall_id",
        "nonfoil",
        "foil",
    ]


def test_row_for_card(id_to_card: Dict[UUID, ScryCard]) -> None:
    card = id_to_card[TEST_CARD_ID]
    card_counts = {counts.CountType.NONFOIL: 3, counts.CountType.FOIL: 5}
    csv_row = csv.row_for_card(card, card_counts)
    assert csv_row == {
        "set": "PHOP",
        "name": "Stairs to Infinity",
        "collector_number": "P1",
        "scryfall_id": TEST_CARD_ID,
        "nonfoil": 3,
        "foil": 5,
    }


def test_rows_for_cards_verbose(oracle: Oracle) -> None:
    card_counts: ScryfallCardCount = {TEST_CARD_ID: {counts.CountType.NONFOIL: 3}}
    collection = MagicCollection(oracle=oracle, counts=card_counts)
    rows = csv.rows_for_cards(collection, True)
    assert list(rows) == [
        {
            "set": "PHOP",
            "name": "Stairs to Infinity",
            "collector_number": "P1",
            "scryfall_id": TEST_CARD_ID,
            "nonfoil": 3,
        },
        {
            "collector_number": "41",
            "name": "Tazeem",
            "scryfall_id": UUID("76e5383d-ac12-4abc-aa30-15e99ded2d6f"),
            "set": "PHOP",
        },
        {
            "set": "PMBS",
            "name": "Hero of Bladehold",
            "collector_number": "8★",
            "scryfall_id": UUID("8829efa0-498a-43ca-91aa-f9caeeafe298"),
        },
        {
            "set": "PMBS",
            "name": "Black Sun's Zenith",
            "collector_number": "39",
            "scryfall_id": UUID("dd88131a-2811-4a1f-bb9a-c82e12c1493b"),
        },
    ]


def test_rows_for_cards_terse(oracle: Oracle) -> None:
    card_counts: counts.ScryfallCardCount = {
        TEST_CARD_ID: {counts.CountType.NONFOIL: 3}
    }
    collection = MagicCollection(oracle=oracle, counts=card_counts)
    rows = csv.rows_for_cards(collection, False)
    assert list(rows) == [
        {
            "set": "PHOP",
            "name": "Stairs to Infinity",
            "collector_number": "P1",
            "scryfall_id": TEST_CARD_ID,
            "nonfoil": 3,
        }
    ]


def test_write_verbose(snapshot: Snapshot, oracle: Oracle, tmp_path: Path) -> None:
    csv_path = tmp_path / "outfile.csv"
    card_counts: ScryfallCardCount = {
        TEST_CARD_ID: {counts.CountType.NONFOIL: 3, counts.CountType.FOIL: 7}
    }
    collection = MagicCollection(oracle=oracle, counts=card_counts)
    serializer = csv.CsvFullDialect()
    serializer.write(csv_path, collection)
    with csv_path.open("rt", encoding="utf-8") as csv_file:
        snapshot.assert_match(csv_file.read(), "collection.csv")


def test_write_terse(snapshot: Snapshot, oracle: Oracle, tmp_path: Path) -> None:
    csv_path = tmp_path / "outfile.csv"
    card_counts: counts.ScryfallCardCount = {
        TEST_CARD_ID: {counts.CountType.NONFOIL: 3}
    }
    collection = MagicCollection(oracle=oracle, counts=card_counts)

    serializer = csv.CsvTerseDialect()
    serializer.write(csv_path, collection)
    with csv_path.open("rt", encoding="utf-8") as csv_file:
        snapshot.assert_match(csv_file.read(), "collection.csv")


def test_read(oracle: Oracle, tmp_path: Path) -> None:
    csv_path = tmp_path / "infile.csv"
    with csv_path.open("wt", encoding="utf-8") as csv_file:
        csv_file.write(
            textwrap.dedent(
                """\
                set,name,collector_number,scryfall_id,nonfoil,foil
                PHOP,Stairs to Infinity,P1,57f25ead-b3ec-4c40-972d-d750ed2f5319,3,7
                """
            )
        )
    serializer = csv.CsvFullDialect()
    collection = serializer.read(csv_path, oracle)
    assert collection.counts == {
        TEST_CARD_ID: {counts.CountType.NONFOIL: 3, counts.CountType.FOIL: 7}
    }



"""Tests for mtg_ssm.manager module."""
# pylint: disable=redefined-outer-name

import argparse as ap
import textwrap
from pathlib import Path

import freezegun
import pytest
from _pytest.monkeypatch import MonkeyPatch

import mtg_ssm.scryfall.fetcher
from mtg_ssm import ssm
from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.containers.indexes import Oracle
from mtg_ssm.scryfall.models import ScryCardLayout, ScrySetType


@pytest.fixture(scope="session")
def oracle(scryfall_data: ScryfallDataSet) -> Oracle:
    """Fixture for Oracle over only MMA and PMBS"""
    accepted_sets = {"mma", "pmbs"}
    scryfall_data2 = ScryfallDataSet(
        sets=[s for s in scryfall_data.sets if s.code in accepted_sets],
        cards=[c for c in scryfall_data.cards if c.set in accepted_sets],
        migrations=[],
    )
    return Oracle(scryfall_data2)


@pytest.fixture(autouse=True)
def patch_scryfetch(monkeypatch: MonkeyPatch, oracle: Oracle) -> None:
    """Fixture to monkeypatch build_card_db for testing."""

    def mock_scryfetch() -> Oracle:
        """Build card_db method that always returns fixed value."""
        return oracle

    monkeypatch.setattr(mtg_ssm.scryfall.fetcher, "scryfetch", mock_scryfetch)


@pytest.mark.parametrize(
    "cmdline, expected",
    [
        (
            "create testfilename",
            ap.Namespace(
                action="create",
                func=ssm.create_cmd,
                collection=Path("testfilename"),
                dialect={},
                include_digital=False,
                include_foreign_only=False,
                separate_promos=False,
                exclude_set_types={ScrySetType.TOKEN, ScrySetType.MEMORABILIA},
                exclude_card_layouts={
                    ScryCardLayout.ART_SERIES,
                    ScryCardLayout.DOUBLE_FACED_TOKEN,
                    ScryCardLayout.EMBLEM,
                    ScryCardLayout.TOKEN,
                },
            ),
        ),
        (
            "--include-digital create testfilename",
            ap.Namespace(
                action="create",
                func=ssm.create_cmd,
                collection=Path("testfilename"),
                dialect={},
                include_digital=True,
                include_foreign_only=False,
                separate_promos=False,
                exclude_set_types={ScrySetType.TOKEN, ScrySetType.MEMORABILIA},
                exclude_card_layouts={
                    ScryCardLayout.ART_SERIES,
                    ScryCardLayout.DOUBLE_FACED_TOKEN,
                    ScryCardLayout.EMBLEM,
                    ScryCardLayout.TOKEN,
                },
            ),
        ),
        (
            "--dialect csv terse create testfilename",
            ap.Namespace(
                action="create",
                func=ssm.create_cmd,
                collection=Path("testfilename"),
                dialect={"csv": "terse"},
                include_digital=False,
                include_foreign_only=False,
                separate_promos=False,
                exclude_set_types={ScrySetType.TOKEN, ScrySetType.MEMORABILIA},
                exclude_card_layouts={
                    ScryCardLayout.ART_SERIES,
                    ScryCardLayout.DOUBLE_FACED_TOKEN,
                    ScryCardLayout.EMBLEM,
                    ScryCardLayout.TOKEN,
                },
            ),
        ),
        (
            "--exclude-set-types=token,memorabilia,vanguard create testfilename",
            ap.Namespace(
                action="create",
                func=ssm.create_cmd,
                collection=Path("testfilename"),
                dialect={},
                include_digital=False,
                include_foreign_only=False,
                separate_promos=False,
                exclude_set_types={
                    ScrySetType.TOKEN,
                    ScrySetType.MEMORABILIA,
                    ScrySetType.VANGUARD,
                },
                exclude_card_layouts={
                    ScryCardLayout.ART_SERIES,
                    ScryCardLayout.DOUBLE_FACED_TOKEN,
                    ScryCardLayout.EMBLEM,
                    ScryCardLayout.TOKEN,
                },
            ),
        ),
        (
            "update testfilename",
            ap.Namespace(
                action="update",
                func=ssm.update_cmd,
                collection=Path("testfilename"),
                dialect={},
                include_digital=False,
                include_foreign_only=False,
                separate_promos=False,
                exclude_set_types={ScrySetType.TOKEN, ScrySetType.MEMORABILIA},
                exclude_card_layouts={
                    ScryCardLayout.ART_SERIES,
                    ScryCardLayout.DOUBLE_FACED_TOKEN,
                    ScryCardLayout.EMBLEM,
                    ScryCardLayout.TOKEN,
                },
            ),
        ),
        (
            "merge testfilename otherfile1",
            ap.Namespace(
                action="merge",
                func=ssm.merge_cmd,
                collection=Path("testfilename"),
                imports=[Path("otherfile1")],
                dialect={},
                include_digital=False,
                include_foreign_only=False,
                separate_promos=False,
                exclude_set_types={ScrySetType.TOKEN, ScrySetType.MEMORABILIA},
                exclude_card_layouts={
                    ScryCardLayout.ART_SERIES,
                    ScryCardLayout.DOUBLE_FACED_TOKEN,
                    ScryCardLayout.EMBLEM,
                    ScryCardLayout.TOKEN,
                },
            ),
        ),
        (
            "merge testfilename otherfile1 otherfile2 otherfile3",
            ap.Namespace(
                action="merge",
                func=ssm.merge_cmd,
                collection=Path("testfilename"),
                imports=[Path("otherfile1"), Path("otherfile2"), Path("otherfile3")],
                dialect={},
                include_digital=False,
                include_foreign_only=False,
                separate_promos=False,
                exclude_set_types={ScrySetType.TOKEN, ScrySetType.MEMORABILIA},
                exclude_card_layouts={
                    ScryCardLayout.ART_SERIES,
                    ScryCardLayout.DOUBLE_FACED_TOKEN,
                    ScryCardLayout.EMBLEM,
                    ScryCardLayout.TOKEN,
                },
            ),
        ),
        (
            "diff file1 file2 file3",
            ap.Namespace(
                action="diff",
                func=ssm.diff_cmd,
                output=Path("file3"),
                left=Path("file1"),
                right=Path("file2"),
                dialect={},
                include_digital=False,
                include_foreign_only=False,
                separate_promos=False,
                exclude_set_types={ScrySetType.TOKEN, ScrySetType.MEMORABILIA},
                exclude_card_layouts={
                    ScryCardLayout.ART_SERIES,
                    ScryCardLayout.DOUBLE_FACED_TOKEN,
                    ScryCardLayout.EMBLEM,
                    ScryCardLayout.TOKEN,
                },
            ),
        ),
    ],
)
def test_get_args(cmdline: str, expected: ap.Namespace) -> None:
    assert ssm.get_args(args=cmdline.split()) == expected


def test_create_cmd(tmp_path: Path, oracle: Oracle) -> None:
    coll_path = tmp_path / "collection.csv"

    args = ap.Namespace(collection=coll_path, dialect={})
    ssm.create_cmd(args, oracle)

    assert coll_path.read_text() == textwrap.dedent(
        """\
        set,name,collector_number,scryfall_id,nonfoil,foil
        PMBS,Hero of Bladehold,8★,8829efa0-498a-43ca-91aa-f9caeeafe298,,
        PMBS,Black Sun's Zenith,39,dd88131a-2811-4a1f-bb9a-c82e12c1493b,,
        MMA,Thallid,167,69d20d28-76e9-4e6e-95c3-f88c51dfabfd,,
        """
    )


@freezegun.freeze_time("2015-06-28 01:02:03")
def test_update_cmd(tmp_path: Path, oracle: Oracle) -> None:
    work_path = tmp_path / "work"
    work_path.mkdir()
    coll_path = work_path / "collection.csv"
    expected_backup_path = work_path / "collection.20150628_010203.csv"

    coll_path.write_text(
        textwrap.dedent(
            """\
            scryfall_id,nonfoil,foil
            69d20d28-76e9-4e6e-95c3-f88c51dfabfd,4,9
            """
        )
    )

    args = ap.Namespace(collection=coll_path, dialect={})
    ssm.update_cmd(args, oracle)

    assert set(work_path.iterdir()) == {coll_path, expected_backup_path}
    assert coll_path.read_text() == textwrap.dedent(
        """\
        set,name,collector_number,scryfall_id,nonfoil,foil
        PMBS,Hero of Bladehold,8★,8829efa0-498a-43ca-91aa-f9caeeafe298,,
        PMBS,Black Sun's Zenith,39,dd88131a-2811-4a1f-bb9a-c82e12c1493b,,
        MMA,Thallid,167,69d20d28-76e9-4e6e-95c3-f88c51dfabfd,4,9
        """
    )
    assert expected_backup_path.read_text() == textwrap.dedent(
        """\
        scryfall_id,nonfoil,foil
        69d20d28-76e9-4e6e-95c3-f88c51dfabfd,4,9
        """
    )


def test_merge_cmd_new(tmp_path: Path, oracle: Oracle) -> None:
    work_path = tmp_path / "work"
    work_path.mkdir()
    coll_path = work_path / "collection.csv"
    import_path = work_path / "import.csv"

    import_path.write_text(
        textwrap.dedent(
            """\
            scryfall_id,nonfoil,foil
            69d20d28-76e9-4e6e-95c3-f88c51dfabfd,4,9
            """
        )
    )

    args = ap.Namespace(collection=coll_path, imports=[import_path], dialect={})
    ssm.merge_cmd(args, oracle)

    assert set(work_path.iterdir()) == {coll_path, import_path}
    assert coll_path.read_text() == textwrap.dedent(
        """\
        set,name,collector_number,scryfall_id,nonfoil,foil
        PMBS,Hero of Bladehold,8★,8829efa0-498a-43ca-91aa-f9caeeafe298,,
        PMBS,Black Sun's Zenith,39,dd88131a-2811-4a1f-bb9a-c82e12c1493b,,
        MMA,Thallid,167,69d20d28-76e9-4e6e-95c3-f88c51dfabfd,4,9
        """
    )
    assert import_path.read_text() == textwrap.dedent(
        """\
        scryfall_id,nonfoil,foil
        69d20d28-76e9-4e6e-95c3-f88c51dfabfd,4,9
        """
    )


@freezegun.freeze_time("2015-06-28 04:05:06")
def test_merge_cmd_existing(tmp_path: Path, oracle: Oracle) -> None:
    work_path = tmp_path / "work"
    work_path.mkdir()
    coll_path = work_path / "collection.csv"
    import_path = work_path / "import.csv"
    expected_backup_path = work_path / "collection.20150628_040506.csv"

    coll_path.write_text(
        textwrap.dedent(
            """\
            scryfall_id,nonfoil,foil
            69d20d28-76e9-4e6e-95c3-f88c51dfabfd,1,3
            """
        )
    )
    import_path.write_text(
        textwrap.dedent(
            """\
            scryfall_id,nonfoil,foil
            69d20d28-76e9-4e6e-95c3-f88c51dfabfd,5,7
            """
        )
    )

    args = ap.Namespace(collection=coll_path, imports=[import_path], dialect={})
    ssm.merge_cmd(args, oracle)

    assert set(work_path.iterdir()) == {coll_path, import_path, expected_backup_path}
    assert coll_path.read_text() == textwrap.dedent(
        """\
        set,name,collector_number,scryfall_id,nonfoil,foil
        PMBS,Hero of Bladehold,8★,8829efa0-498a-43ca-91aa-f9caeeafe298,,
        PMBS,Black Sun's Zenith,39,dd88131a-2811-4a1f-bb9a-c82e12c1493b,,
        MMA,Thallid,167,69d20d28-76e9-4e6e-95c3-f88c51dfabfd,6,10
        """
    )
    assert import_path.read_text() == textwrap.dedent(
        """\
        scryfall_id,nonfoil,foil
        69d20d28-76e9-4e6e-95c3-f88c51dfabfd,5,7
        """
    )
    assert expected_backup_path.read_text() == textwrap.dedent(
        """\
        scryfall_id,nonfoil,foil
        69d20d28-76e9-4e6e-95c3-f88c51dfabfd,1,3
        """
    )


@freezegun.freeze_time("2015-06-28 08:09:10")
def test_merge_cmd_multiple(tmp_path: Path, oracle: Oracle) -> None:
    work_path = tmp_path / "work"
    work_path.mkdir()
    coll_path = work_path / "collection.csv"
    import_path1 = work_path / "import1.csv"
    import_path2 = work_path / "import2.csv"
    expected_backup_path = work_path / "collection.20150628_080910.csv"

    coll_path.write_text(
        textwrap.dedent(
            """\
            scryfall_id,nonfoil,foil
            69d20d28-76e9-4e6e-95c3-f88c51dfabfd,1,3
            """
        )
    )
    import_path1.write_text(
        textwrap.dedent(
            """\
            scryfall_id,nonfoil,foil
            69d20d28-76e9-4e6e-95c3-f88c51dfabfd,5,7
            """
        )
    )
    import_path2.write_text(
        textwrap.dedent(
            """\
            scryfall_id,nonfoil,foil
            dd88131a-2811-4a1f-bb9a-c82e12c1493b,19,23
            """
        )
    )

    args = ap.Namespace(
        collection=coll_path, imports=[import_path1, import_path2], dialect={}
    )
    ssm.merge_cmd(args, oracle)

    assert set(work_path.iterdir()) == {
        coll_path,
        import_path1,
        import_path2,
        expected_backup_path,
    }
    assert coll_path.read_text() == textwrap.dedent(
        """\
        set,name,collector_number,scryfall_id,nonfoil,foil
        PMBS,Hero of Bladehold,8★,8829efa0-498a-43ca-91aa-f9caeeafe298,,
        PMBS,Black Sun's Zenith,39,dd88131a-2811-4a1f-bb9a-c82e12c1493b,19,23
        MMA,Thallid,167,69d20d28-76e9-4e6e-95c3-f88c51dfabfd,6,10
        """
    )
    assert import_path1.read_text() == textwrap.dedent(
        """\
        scryfall_id,nonfoil,foil
        69d20d28-76e9-4e6e-95c3-f88c51dfabfd,5,7
        """
    )
    assert import_path2.read_text() == textwrap.dedent(
        """\
        scryfall_id,nonfoil,foil
        dd88131a-2811-4a1f-bb9a-c82e12c1493b,19,23
        """
    )
    assert expected_backup_path.read_text() == textwrap.dedent(
        """\
        scryfall_id,nonfoil,foil
        69d20d28-76e9-4e6e-95c3-f88c51dfabfd,1,3
        """
    )


def test_diff_cmd(tmp_path: Path, oracle: Oracle) -> None:
    work_path = tmp_path / "work"
    work_path.mkdir()
    left_path = work_path / "left.csv"
    right_path = work_path / "right.csv"
    out_path = work_path / "out.csv"

    left_path.write_text(
        textwrap.dedent(
            """\
            scryfall_id,nonfoil,foil
            69d20d28-76e9-4e6e-95c3-f88c51dfabfd,8,
            """
        )
    )
    right_path.write_text(
        textwrap.dedent(
            """\
            scryfall_id,nonfoil,foil
            dd88131a-2811-4a1f-bb9a-c82e12c1493b,4,
            69d20d28-76e9-4e6e-95c3-f88c51dfabfd,1,3
            """
        )
    )

    args = ap.Namespace(output=out_path, left=left_path, right=right_path, dialect={})
    ssm.diff_cmd(args, oracle)
    assert set(work_path.iterdir()) == {left_path, right_path, out_path}
    assert out_path.read_text() == textwrap.dedent(
        """\
        set,name,collector_number,scryfall_id,nonfoil,foil
        PMBS,Hero of Bladehold,8★,8829efa0-498a-43ca-91aa-f9caeeafe298,,
        PMBS,Black Sun's Zenith,39,dd88131a-2811-4a1f-bb9a-c82e12c1493b,-4,
        MMA,Thallid,167,69d20d28-76e9-4e6e-95c3-f88c51dfabfd,7,-3
        """
    )



"""Helpers for tracking collection counts."""

import collections
import enum
from typing import Any, Dict, Iterable, MutableMapping
from uuid import UUID

from mtg_ssm.containers import legacy
from mtg_ssm.containers.indexes import Oracle


class Error(Exception):
    """Base exception for this module."""


class CardNotFoundError(Error):
    """Raised when we attempt to add a card to a count but cannot find it in the oracle."""


class CountType(str, enum.Enum):
    """Enum for possible card printing types (nonfoil, foil)."""

    NONFOIL = "nonfoil"
    FOIL = "foil"


ScryfallCardCount = Dict[UUID, MutableMapping[CountType, int]]
"""Mapping from scryfall id to card printing type to count."""


def aggregate_card_counts(
    card_rows: Iterable[Dict[str, Any]], oracle: Oracle
) -> ScryfallCardCount:
    """Extract card counts from card rows."""
    card_counts: ScryfallCardCount = {}
    for card_row in card_rows:
        if "scryfall_id" not in card_row:
            card_row = legacy.coerce_row(card_row, oracle)
        if not card_row:
            continue
        scryfall_id = card_row["scryfall_id"]
        if not isinstance(scryfall_id, UUID):
            scryfall_id = UUID(scryfall_id)
        counts = card_counts.get(scryfall_id, {})
        for count_type in CountType:
            value = int(card_row.get(count_type.value) or 0)
            if value:
                counts[count_type] = value + counts.get(count_type, 0)
        if counts:
            while scryfall_id in oracle.index.migrate_old_id_to_new_id:
                scryfall_id = oracle.index.migrate_old_id_to_new_id[scryfall_id]
            if scryfall_id not in oracle.index.id_to_card:
                raise CardNotFoundError(
                    f"Found counts for card={scryfall_id} not found scryfall data"
                )
            card_counts[scryfall_id] = counts
    return card_counts


def merge_card_counts(*card_counts_args: ScryfallCardCount) -> ScryfallCardCount:
    """Merge two sets of card_counts."""
    merged_counts: ScryfallCardCount = collections.defaultdict(collections.Counter)
    for card_counts in card_counts_args:
        for card_id, counts in card_counts.items():
            merged_counts[card_id].update(counts)
    return dict(merged_counts)


def diff_card_counts(
    left: ScryfallCardCount, right: ScryfallCardCount
) -> ScryfallCardCount:
    """Subtract right print counts from left print counts."""
    diffed_counts: ScryfallCardCount = collections.defaultdict(dict)
    for card_id in left.keys() | right.keys():
        left_counts = left.get(card_id, {})
        right_counts = right.get(card_id, {})
        card_counts = {
            k: left_counts.get(k, 0) - right_counts.get(k, 0)
            for k in left_counts.keys() | right_counts.keys()
        }
        card_counts = {k: v for k, v in card_counts.items() if v}
        if card_counts:
            diffed_counts[card_id] = card_counts
    return diffed_counts
